@@ -28,30 +28,43 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 #ifndef EV_STANDALONE
 # include "config.h"
 
 # if HAVE_CLOCK_GETTIME
-#  define EV_USE_MONOTONIC 1
-#  define EV_USE_REALTIME  1
+#  ifndef EV_USE_MONOTONIC
+#   define EV_USE_MONOTONIC 1
+#  endif
+#  ifndef EV_USE_REALTIME
+#   define EV_USE_REALTIME  1
+#  endif
 # endif
 
-# if HAVE_SELECT && HAVE_SYS_SELECT_H
+# if HAVE_SELECT && HAVE_SYS_SELECT_H && !defined (EV_USE_SELECT)
 #  define EV_USE_SELECT 1
 # endif
 
-# if HAVE_POLL && HAVE_POLL_H
+# if HAVE_POLL && HAVE_POLL_H && !defined (EV_USE_POLL)
 #  define EV_USE_POLL 1
 # endif
 
-# if HAVE_EPOLL && HAVE_EPOLL_CTL && HAVE_SYS_EPOLL_H
+# if HAVE_EPOLL_CTL && HAVE_SYS_EPOLL_H && !defined (EV_USE_EPOLL)
 #  define EV_USE_EPOLL 1
 # endif
 
-# if HAVE_KQUEUE && HAVE_WORKING_KQUEUE && HAVE_SYS_EVENT_H && HAVE_SYS_QUEUE_H
+# if HAVE_KQUEUE && HAVE_SYS_EVENT_H && HAVE_SYS_QUEUE_H && !defined (EV_USE_KQUEUE)
 #  define EV_USE_KQUEUE 1
 # endif
 
+# if HAVE_PORT_H && HAVE_PORT_CREATE && !defined (EV_USE_PORT)
+#  define EV_USE_PORT 1
+# endif
+
 #endif
 
 #include <math.h>
@@ -68,15 +81,26 @@
 
 #include <signal.h>
 
-#ifndef WIN32
+#ifndef _WIN32
 # include <unistd.h>
 # include <sys/time.h>
 # include <sys/wait.h>
+#else
+# define WIN32_LEAN_AND_MEAN
+# include <windows.h>
+# ifndef EV_SELECT_IS_WINSOCKET
+#  define EV_SELECT_IS_WINSOCKET 1
+# endif
 #endif
+
 /**/
 
 #ifndef EV_USE_MONOTONIC
-# define EV_USE_MONOTONIC 1
+# define EV_USE_MONOTONIC 0
+#endif
+
+#ifndef EV_USE_REALTIME
+# define EV_USE_REALTIME 0
 #endif
 
 #ifndef EV_USE_SELECT
@@ -84,7 +108,11 @@
 #endif
 
 #ifndef EV_USE_POLL
-# define EV_USE_POLL 0 /* poll is usually slower than select, and not as well tested */
+# ifdef _WIN32
+#  define EV_USE_POLL 0
+# else
+#  define EV_USE_POLL 1
+# endif
 #endif
 
 #ifndef EV_USE_EPOLL
@@ -95,22 +123,18 @@
 # define EV_USE_KQUEUE 0
 #endif
 
-#ifndef EV_USE_WIN32
-# ifdef WIN32
-#  define EV_USE_WIN32 0 /* it does not exist, use select */
-#  undef EV_USE_SELECT
-#  define EV_USE_SELECT 1
-# else
-#  define EV_USE_WIN32 0
-# endif
-#endif
-
-#ifndef EV_USE_REALTIME
-# define EV_USE_REALTIME 1
+#ifndef EV_USE_PORT
+# define EV_USE_PORT 0
 #endif
 
 /**/
 
+/* darwin simply cannot be helped */
+#ifdef __APPLE__
+# undef EV_USE_POLL
+# undef EV_USE_KQUEUE
+#endif
+
 #ifndef CLOCK_MONOTONIC
 # undef EV_USE_MONOTONIC
 # define EV_USE_MONOTONIC 0
@@ -121,14 +145,22 @@
 # define EV_USE_REALTIME 0
 #endif
 
+#if EV_SELECT_IS_WINSOCKET
+# include <winsock.h>
+#endif
+
 /**/
 
 #define MIN_TIMEJUMP  1. /* minimum timejump that gets detected (if monotonic clock available) */
-#define MAX_BLOCKTIME 59.731 /* never wait longer than this time (to detect time jumps) */
+#define MAX_BLOCKTIME 59.743 /* never wait longer than this time (to detect time jumps) */
 #define PID_HASHSIZE  16 /* size of pid hash table, must be power of two */
-/*#define CLEANUP_INTERVAL 300. /* how often to try to free memory and re-check fds */
+/*#define CLEANUP_INTERVAL (MAX_BLOCKTIME * 5.) /* how often to try to free memory and re-check fds */
 
-#include "ev.h"
+#ifdef EV_H
+# include EV_H
+#else
+# include "ev.h"
+#endif
 
 #if __GNUC__ >= 3
 # define expect(expr,value)         __builtin_expect ((expr),(value))
@@ -144,13 +176,18 @@
 #define NUMPRI    (EV_MAXPRI - EV_MINPRI + 1)
 #define ABSPRI(w) ((w)->priority - EV_MINPRI)
 
+#define EMPTY0      /* required for microsofts broken pseudo-c compiler */
+#define EMPTY2(a,b) /* used to suppress some warnings */
+
 typedef struct ev_watcher *W;
 typedef struct ev_watcher_list *WL;
 typedef struct ev_watcher_time *WT;
 
 static int have_monotonic; /* did clock_gettime (CLOCK_MONOTONIC) work? */
 
-#include "ev_win32.c"
+#ifdef _WIN32
+# include "ev_win32.c"
+#endif
 
 /*****************************************************************************/
 
@@ -207,6 +244,9 @@ typedef struct
   WL head;
   unsigned char events;
   unsigned char reify;
+#if EV_SELECT_IS_WINSOCKET
+  SOCKET handle;
+#endif
 } ANFD;
 
 typedef struct
@@ -217,25 +257,33 @@ typedef struct
 
 #if EV_MULTIPLICITY
 
-struct ev_loop
-{
-# define VAR(name,decl) decl;
-# include "ev_vars.h"
-};
-# undef VAR
-# include "ev_wrap.h"
+  struct ev_loop
+  {
+    ev_tstamp ev_rt_now;
+    #define ev_rt_now ((loop)->ev_rt_now)
+    #define VAR(name,decl) decl;
+      #include "ev_vars.h"
+    #undef VAR
+  };
+  #include "ev_wrap.h"
+
+  static struct ev_loop default_loop_struct;
+  struct ev_loop *ev_default_loop_ptr;
 
 #else
 
-# define VAR(name,decl) static decl;
-# include "ev_vars.h"
-# undef VAR
+  ev_tstamp ev_rt_now;
+  #define VAR(name,decl) static decl;
+    #include "ev_vars.h"
+  #undef VAR
+
+  static int ev_default_loop_ptr;
 
 #endif
 
 /*****************************************************************************/
 
-inline ev_tstamp
+ev_tstamp
 ev_time (void)
 {
 #if EV_USE_REALTIME
@@ -264,13 +312,15 @@ get_clock (void)
   return ev_time ();
 }
 
+#if EV_MULTIPLICITY
 ev_tstamp
 ev_now (EV_P)
 {
-  return rt_now;
+  return ev_rt_now;
 }
+#endif
 
-#define array_roundsize(type,n) ((n) | 4 & ~3)
+#define array_roundsize(type,n) (((n) | 4) & ~3)
 
 #define array_needsize(type,base,cur,cnt,init)			\
   if (expect_false ((cnt) > cur))				\
@@ -295,11 +345,6 @@ ev_now (EV_P)
       fprintf (stderr, "slimmed down " # stem " to %d\n", stem ## max);/*D*/\
     }
 
-/* microsoft's pseudo-c is quite far from C as the rest of the world and the standard knows it */
-/* bringing us everlasting joy in form of stupid extra macros that are not required in C */
-#define array_free_microshit(stem) \
-  ev_free (stem ## s); stem ## cnt = stem ## max = 0;
-
 #define array_free(stem, idx) \
   ev_free (stem ## s idx); stem ## cnt idx = stem ## max idx = 0;
 
@@ -318,19 +363,21 @@ anfds_init (ANFD *base, int count)
     }
 }
 
-static void
-event (EV_P_ W w, int events)
+void
+ev_feed_event (EV_P_ void *w, int revents)
 {
-  if (w->pending)
+  W w_ = (W)w;
+
+  if (w_->pending)
     {
-      pendings [ABSPRI (w)][w->pending - 1].events |= events;
+      pendings [ABSPRI (w_)][w_->pending - 1].events |= revents;
       return;
     }
 
-  w->pending = ++pendingcnt [ABSPRI (w)];
-  array_needsize (ANPENDING, pendings [ABSPRI (w)], pendingmax [ABSPRI (w)], pendingcnt [ABSPRI (w)], (void));
-  pendings [ABSPRI (w)][w->pending - 1].w      = w;
-  pendings [ABSPRI (w)][w->pending - 1].events = events;
+  w_->pending = ++pendingcnt [ABSPRI (w_)];
+  array_needsize (ANPENDING, pendings [ABSPRI (w_)], pendingmax [ABSPRI (w_)], pendingcnt [ABSPRI (w_)], EMPTY2);
+  pendings [ABSPRI (w_)][w_->pending - 1].w      = w_;
+  pendings [ABSPRI (w_)][w_->pending - 1].events = revents;
 }
 
 static void
@@ -339,24 +386,30 @@ queue_events (EV_P_ W *events, int eventcnt, int type)
   int i;
 
   for (i = 0; i < eventcnt; ++i)
-    event (EV_A_ events [i], type);
+    ev_feed_event (EV_A_ events [i], type);
 }
 
-static void
-fd_event (EV_P_ int fd, int events)
+inline void
+fd_event (EV_P_ int fd, int revents)
 {
   ANFD *anfd = anfds + fd;
   struct ev_io *w;
 
   for (w = (struct ev_io *)anfd->head; w; w = (struct ev_io *)((WL)w)->next)
     {
-      int ev = w->events & events;
+      int ev = w->events & revents;
 
       if (ev)
-        event (EV_A_ (W)w, ev);
+        ev_feed_event (EV_A_ (W)w, ev);
     }
 }
 
+void
+ev_feed_fd_event (EV_P_ int fd, int revents)
+{
+  fd_event (EV_A_ fd, revents);
+}
+
 /*****************************************************************************/
 
 static void
@@ -375,6 +428,15 @@ fd_reify (EV_P)
       for (w = (struct ev_io *)anfd->head; w; w = (struct ev_io *)((WL)w)->next)
         events |= w->events;
 
+#if EV_SELECT_IS_WINSOCKET
+      if (events)
+        {
+          unsigned long argp;
+          anfd->handle = _get_osfhandle (fd);
+          assert (("libev only supports socket fds in this configuration", ioctlsocket (anfd->handle, FIONREAD, &argp) == 0));
+        }
+#endif
+
       anfd->reify = 0;
 
       method_modify (EV_A_ fd, anfd->events, events);
@@ -393,7 +455,7 @@ fd_change (EV_P_ int fd)
   anfds [fd].reify = 1;
 
   ++fdchangecnt;
-  array_needsize (int, fdchanges, fdchangemax, fdchangecnt, (void));
+  array_needsize (int, fdchanges, fdchangemax, fdchangecnt, EMPTY2);
   fdchanges [fdchangecnt - 1] = fd;
 }
 
@@ -405,15 +467,15 @@ fd_kill (EV_P_ int fd)
   while ((w = (struct ev_io *)anfds [fd].head))
     {
       ev_io_stop (EV_A_ w);
-      event (EV_A_ (W)w, EV_ERROR | EV_READ | EV_WRITE);
+      ev_feed_event (EV_A_ (W)w, EV_ERROR | EV_READ | EV_WRITE);
     }
 }
 
 static int
 fd_valid (int fd)
 {
-#ifdef WIN32
-  return !!win32_get_osfhandle (fd);
+#ifdef _WIN32
+  return _get_osfhandle (fd) != -1;
 #else
   return fcntl (fd, F_GETFD) != -1;
 #endif
@@ -503,6 +565,13 @@ downheap (WT *heap, int N, int k)
   ((W)heap [k])->active = k + 1;
 }
 
+inline void
+adjustheap (WT *heap, int N, int k)
+{
+  upheap (heap, k);
+  downheap (heap, N, k);
+}
+
 /*****************************************************************************/
 
 typedef struct
@@ -533,7 +602,7 @@ signals_init (ANSIG *base, int count)
 static void
 sighandler (int signum)
 {
-#if WIN32
+#if _WIN32
   signal (signum, sighandler);
 #endif
 
@@ -548,10 +617,29 @@ sighandler (int signum)
     }
 }
 
+void
+ev_feed_signal_event (EV_P_ int signum)
+{
+  WL w;
+
+#if EV_MULTIPLICITY
+  assert (("feeding signal events is only supported in the default loop", loop == ev_default_loop_ptr));
+#endif
+
+  --signum;
+
+  if (signum < 0 || signum >= signalmax)
+    return;
+
+  signals [signum].gotsig = 0;
+
+  for (w = signals [signum].head; w; w = w->next)
+    ev_feed_event (EV_A_ (W)w, EV_SIGNAL);
+}
+
 static void
 sigcb (EV_P_ struct ev_io *iow, int revents)
 {
-  WL w;
   int signum;
 
   read (sigpipe [0], &revents, 1);
@@ -559,25 +647,26 @@ sigcb (EV_P_ struct ev_io *iow, int revents)
 
   for (signum = signalmax; signum--; )
     if (signals [signum].gotsig)
-      {
-        signals [signum].gotsig = 0;
+      ev_feed_signal_event (EV_A_ signum + 1);
+}
 
-        for (w = signals [signum].head; w; w = w->next)
-          event (EV_A_ (W)w, EV_SIGNAL);
-      }
+inline void
+fd_intern (int fd)
+{
+#ifdef _WIN32
+  int arg = 1;
+  ioctlsocket (_get_osfhandle (fd), FIONBIO, &arg);
+#else
+  fcntl (fd, F_SETFD, FD_CLOEXEC);
+  fcntl (fd, F_SETFL, O_NONBLOCK);
+#endif
 }
 
 static void
 siginit (EV_P)
 {
-#ifndef WIN32
-  fcntl (sigpipe [0], F_SETFD, FD_CLOEXEC);
-  fcntl (sigpipe [1], F_SETFD, FD_CLOEXEC);
-
-  /* rather than sort out wether we really need nb, set it */
-  fcntl (sigpipe [0], F_SETFL, O_NONBLOCK);
-  fcntl (sigpipe [1], F_SETFL, O_NONBLOCK);
-#endif
+  fd_intern (sigpipe [0]);
+  fd_intern (sigpipe [1]);
 
   ev_io_set (&sigev, sigpipe [0], EV_READ);
   ev_io_start (EV_A_ &sigev);
@@ -588,7 +677,7 @@ siginit (EV_P)
 
 static struct ev_child *childs [PID_HASHSIZE];
 
-#ifndef WIN32
+#ifndef _WIN32
 
 static struct ev_signal childev;
 
@@ -607,7 +696,7 @@ child_reap (EV_P_ struct ev_signal *sw, int chain, int pid, int status)
         ev_priority (w) = ev_priority (sw); /* need to do it *now* */
         w->rpid         = pid;
         w->rstatus      = status;
-        event (EV_A_ (W)w, EV_CHILD);
+        ev_feed_event (EV_A_ (W)w, EV_CHILD);
       }
 }
 
@@ -619,7 +708,7 @@ childcb (EV_P_ struct ev_signal *sw, int revents)
   if (0 < (pid = waitpid (-1, &status, WNOHANG | WUNTRACED | WCONTINUED)))
     {
       /* make sure we are called again until all childs have been reaped */
-      event (EV_A_ (W)sw, EV_SIGNAL);
+      ev_feed_event (EV_A_ (W)sw, EV_SIGNAL);
 
       child_reap (EV_A_ sw, pid, pid, status);
       child_reap (EV_A_ sw,   0, pid, status); /* this might trigger a watcher twice, but event catches that */
@@ -630,6 +719,9 @@ childcb (EV_P_ struct ev_signal *sw, int revents)
 
 /*****************************************************************************/
 
+#if EV_USE_PORT
+# include "ev_port.c"
+#endif
 #if EV_USE_KQUEUE
 # include "ev_kqueue.c"
 #endif
@@ -659,7 +751,7 @@ ev_version_minor (void)
 static int
 enable_secure (void)
 {
-#ifdef WIN32
+#ifdef _WIN32
   return 0;
 #else
   return getuid () != geteuid ()
@@ -667,14 +759,14 @@ enable_secure (void)
 #endif
 }
 
-int
+unsigned int
 ev_method (EV_P)
 {
   return method;
 }
 
 static void
-loop_init (EV_P_ int methods)
+loop_init (EV_P_ unsigned int flags)
 {
   if (!method)
     {
@@ -686,35 +778,35 @@ loop_init (EV_P_ int methods)
       }
 #endif
 
-      rt_now    = ev_time ();
+      ev_rt_now = ev_time ();
       mn_now    = get_clock ();
       now_floor = mn_now;
-      rtmn_diff = rt_now - mn_now;
+      rtmn_diff = ev_rt_now - mn_now;
 
-      if (methods == EVMETHOD_AUTO)
-        if (!enable_secure () && getenv ("LIBEV_METHODS"))
-          methods = atoi (getenv ("LIBEV_METHODS"));
-        else
-          methods = EVMETHOD_ANY;
+      if (!(flags & EVFLAG_NOENV) && !enable_secure () && getenv ("LIBEV_FLAGS"))
+        flags = atoi (getenv ("LIBEV_FLAGS"));
+
+      if (!(flags & 0x0000ffff))
+        flags |= 0x0000ffff;
 
       method = 0;
-#if EV_USE_WIN32
-      if (!method && (methods & EVMETHOD_WIN32 )) method = win32_init  (EV_A_ methods);
+#if EV_USE_PORT
+      if (!method && (flags & EVMETHOD_PORT  )) method = port_init   (EV_A_ flags);
 #endif
 #if EV_USE_KQUEUE
-      if (!method && (methods & EVMETHOD_KQUEUE)) method = kqueue_init (EV_A_ methods);
+      if (!method && (flags & EVMETHOD_KQUEUE)) method = kqueue_init (EV_A_ flags);
 #endif
 #if EV_USE_EPOLL
-      if (!method && (methods & EVMETHOD_EPOLL )) method = epoll_init  (EV_A_ methods);
+      if (!method && (flags & EVMETHOD_EPOLL )) method = epoll_init  (EV_A_ flags);
 #endif
 #if EV_USE_POLL
-      if (!method && (methods & EVMETHOD_POLL  )) method = poll_init   (EV_A_ methods);
+      if (!method && (flags & EVMETHOD_POLL  )) method = poll_init   (EV_A_ flags);
 #endif
 #if EV_USE_SELECT
-      if (!method && (methods & EVMETHOD_SELECT)) method = select_init (EV_A_ methods);
+      if (!method && (flags & EVMETHOD_SELECT)) method = select_init (EV_A_ flags);
 #endif
 
-      ev_watcher_init (&sigev, sigcb);
+      ev_init (&sigev, sigcb);
       ev_set_priority (&sigev, EV_MAXPRI);
     }
 }
@@ -724,8 +816,8 @@ loop_destroy (EV_P)
 {
   int i;
 
-#if EV_USE_WIN32
-  if (method == EVMETHOD_WIN32 ) win32_destroy  (EV_A);
+#if EV_USE_PORT
+  if (method == EVMETHOD_PORT  ) port_destroy   (EV_A);
 #endif
 #if EV_USE_KQUEUE
   if (method == EVMETHOD_KQUEUE) kqueue_destroy (EV_A);
@@ -744,12 +836,14 @@ loop_destroy (EV_P)
     array_free (pending, [i]);
 
   /* have to use the microsoft-never-gets-it-right macro */
-  array_free_microshit (fdchange);
-  array_free_microshit (timer);
-  array_free_microshit (periodic);
-  array_free_microshit (idle);
-  array_free_microshit (prepare);
-  array_free_microshit (check);
+  array_free (fdchange, EMPTY0);
+  array_free (timer, EMPTY0);
+#if EV_PERIODICS
+  array_free (periodic, EMPTY0);
+#endif
+  array_free (idle, EMPTY0);
+  array_free (prepare, EMPTY0);
+  array_free (check, EMPTY0);
 
   method = 0;
 }
@@ -757,12 +851,15 @@ loop_destroy (EV_P)
 static void
 loop_fork (EV_P)
 {
-#if EV_USE_EPOLL
-  if (method == EVMETHOD_EPOLL ) epoll_fork  (EV_A);
+#if EV_USE_PORT
+  if (method == EVMETHOD_PORT  ) port_fork   (EV_A);
 #endif
 #if EV_USE_KQUEUE
   if (method == EVMETHOD_KQUEUE) kqueue_fork (EV_A);
 #endif
+#if EV_USE_EPOLL
+  if (method == EVMETHOD_EPOLL ) epoll_fork  (EV_A);
+#endif
 
   if (ev_is_active (&sigev))
     {
@@ -784,13 +881,13 @@ loop_fork (EV_P)
 
 #if EV_MULTIPLICITY
 struct ev_loop *
-ev_loop_new (int methods)
+ev_loop_new (unsigned int flags)
 {
   struct ev_loop *loop = (struct ev_loop *)ev_malloc (sizeof (struct ev_loop));
 
   memset (loop, 0, sizeof (struct ev_loop));
 
-  loop_init (EV_A_ methods);
+  loop_init (EV_A_ flags);
 
   if (ev_method (EV_A))
     return loop;
@@ -814,36 +911,32 @@ ev_loop_fork (EV_P)
 #endif
 
 #if EV_MULTIPLICITY
-struct ev_loop default_loop_struct;
-static struct ev_loop *default_loop;
-
 struct ev_loop *
+ev_default_loop_ (unsigned int flags)
 #else
-static int default_loop;
-
 int
+ev_default_loop (unsigned int flags)
 #endif
-ev_default_loop (int methods)
 {
   if (sigpipe [0] == sigpipe [1])
     if (pipe (sigpipe))
       return 0;
 
-  if (!default_loop)
+  if (!ev_default_loop_ptr)
     {
 #if EV_MULTIPLICITY
-      struct ev_loop *loop = default_loop = &default_loop_struct;
+      struct ev_loop *loop = ev_default_loop_ptr = &default_loop_struct;
 #else
-      default_loop = 1;
+      ev_default_loop_ptr = 1;
 #endif
 
-      loop_init (EV_A_ methods);
+      loop_init (EV_A_ flags);
 
       if (ev_method (EV_A))
         {
           siginit (EV_A);
 
-#ifndef WIN32
+#ifndef _WIN32
           ev_signal_init (&childev, childcb, SIGCHLD);
           ev_set_priority (&childev, EV_MAXPRI);
           ev_signal_start (EV_A_ &childev);
@@ -851,20 +944,20 @@ ev_default_loop (int methods)
 #endif
         }
       else
-        default_loop = 0;
+        ev_default_loop_ptr = 0;
     }
 
-  return default_loop;
+  return ev_default_loop_ptr;
 }
 
 void
 ev_default_destroy (void)
 {
 #if EV_MULTIPLICITY
-  struct ev_loop *loop = default_loop;
+  struct ev_loop *loop = ev_default_loop_ptr;
 #endif
 
-#ifndef WIN32
+#ifndef _WIN32
   ev_ref (EV_A); /* child watcher */
   ev_signal_stop (EV_A_ &childev);
 #endif
@@ -882,7 +975,7 @@ void
 ev_default_fork (void)
 {
 #if EV_MULTIPLICITY
-  struct ev_loop *loop = default_loop;
+  struct ev_loop *loop = ev_default_loop_ptr;
 #endif
 
   if (method)
@@ -891,6 +984,18 @@ ev_default_fork (void)
 
 /*****************************************************************************/
 
+static int
+any_pending (EV_P)
+{
+  int pri;
+
+  for (pri = NUMPRI; pri--; )
+    if (pendingcnt [pri])
+      return 1;
+
+  return 0;
+}
+
 static void
 call_pending (EV_P)
 {
@@ -904,7 +1009,7 @@ call_pending (EV_P)
         if (p->w)
           {
             p->w->pending = 0;
-            p->w->cb (EV_A_ p->w, p->events);
+            EV_CB_INVOKE (p->w, p->events);
           }
       }
 }
@@ -922,36 +1027,47 @@ timers_reify (EV_P)
       if (w->repeat)
         {
           assert (("negative ev_timer repeat value found while processing timers", w->repeat > 0.));
-          ((WT)w)->at = mn_now + w->repeat;
+
+          ((WT)w)->at += w->repeat;
+          if (((WT)w)->at < mn_now)
+            ((WT)w)->at = mn_now;
+
           downheap ((WT *)timers, timercnt, 0);
         }
       else
         ev_timer_stop (EV_A_ w); /* nonrepeating: stop timer */
 
-      event (EV_A_ (W)w, EV_TIMEOUT);
+      ev_feed_event (EV_A_ (W)w, EV_TIMEOUT);
     }
 }
 
+#if EV_PERIODICS
 static void
 periodics_reify (EV_P)
 {
-  while (periodiccnt && ((WT)periodics [0])->at <= rt_now)
+  while (periodiccnt && ((WT)periodics [0])->at <= ev_rt_now)
     {
       struct ev_periodic *w = periodics [0];
 
       assert (("inactive timer on periodic heap detected", ev_is_active (w)));
 
       /* first reschedule or stop timer */
-      if (w->interval)
+      if (w->reschedule_cb)
         {
-          ((WT)w)->at += floor ((rt_now - ((WT)w)->at) / w->interval + 1.) * w->interval;
-          assert (("ev_periodic timeout in the past detected while processing timers, negative interval?", ((WT)w)->at > rt_now));
+          ((WT)w)->at = w->reschedule_cb (w, ev_rt_now + 0.0001);
+          assert (("ev_periodic reschedule callback returned time in the past", ((WT)w)->at > ev_rt_now));
+          downheap ((WT *)periodics, periodiccnt, 0);
+        }
+      else if (w->interval)
+        {
+          ((WT)w)->at += floor ((ev_rt_now - ((WT)w)->at) / w->interval + 1.) * w->interval;
+          assert (("ev_periodic timeout in the past detected while processing timers, negative interval?", ((WT)w)->at > ev_rt_now));
           downheap ((WT *)periodics, periodiccnt, 0);
         }
       else
         ev_periodic_stop (EV_A_ w); /* nonrepeating: stop timer */
 
-      event (EV_A_ (W)w, EV_PERIODIC);
+      ev_feed_event (EV_A_ (W)w, EV_PERIODIC);
     }
 }
 
@@ -965,20 +1081,17 @@ periodics_reschedule (EV_P)
     {
       struct ev_periodic *w = periodics [i];
 
-      if (w->interval)
-        {
-          ev_tstamp diff = ceil ((rt_now - ((WT)w)->at) / w->interval) * w->interval;
-
-          if (fabs (diff) >= 1e-4)
-            {
-              ev_periodic_stop (EV_A_ w);
-              ev_periodic_start (EV_A_ w);
-
-              i = 0; /* restart loop, inefficient, but time jumps should be rare */
-            }
-        }
+      if (w->reschedule_cb)
+        ((WT)w)->at = w->reschedule_cb (w, ev_rt_now);
+      else if (w->interval)
+        ((WT)w)->at += ceil ((ev_rt_now - ((WT)w)->at) / w->interval) * w->interval;
     }
+
+  /* now rebuild the heap */
+  for (i = periodiccnt >> 1; i--; )
+    downheap ((WT *)periodics, periodiccnt, i);
 }
+#endif
 
 inline int
 time_update_monotonic (EV_P)
@@ -987,13 +1100,13 @@ time_update_monotonic (EV_P)
 
   if (expect_true (mn_now - now_floor < MIN_TIMEJUMP * .5))
     {
-      rt_now = rtmn_diff + mn_now;
+      ev_rt_now = rtmn_diff + mn_now;
       return 0;
     }
   else
     {
       now_floor = mn_now;
-      rt_now = ev_time ();
+      ev_rt_now = ev_time ();
       return 1;
     }
 }
@@ -1012,17 +1125,19 @@ time_update (EV_P)
 
           for (i = 4; --i; ) /* loop a few times, before making important decisions */
             {
-              rtmn_diff = rt_now - mn_now;
+              rtmn_diff = ev_rt_now - mn_now;
 
               if (fabs (odiff - rtmn_diff) < MIN_TIMEJUMP)
                 return; /* all is well */
 
-              rt_now    = ev_time ();
+              ev_rt_now = ev_time ();
               mn_now    = get_clock ();
               now_floor = mn_now;
             }
 
+# if EV_PERIODICS
           periodics_reschedule (EV_A);
+# endif
           /* no timer adjustment, as the monotonic clock doesn't jump */
           /* timers_reschedule (EV_A_ rtmn_diff - odiff) */
         }
@@ -1030,18 +1145,20 @@ time_update (EV_P)
   else
 #endif
     {
-      rt_now = ev_time ();
+      ev_rt_now = ev_time ();
 
-      if (expect_false (mn_now > rt_now || mn_now < rt_now - MAX_BLOCKTIME - MIN_TIMEJUMP))
+      if (expect_false (mn_now > ev_rt_now || mn_now < ev_rt_now - MAX_BLOCKTIME - MIN_TIMEJUMP))
         {
+#if EV_PERIODICS
           periodics_reschedule (EV_A);
+#endif
 
           /* adjust timers. this is easy, as the offset is the same for all */
           for (i = 0; i < timercnt; ++i)
-            ((WT)timers [i])->at += rt_now - mn_now;
+            ((WT)timers [i])->at += ev_rt_now - mn_now;
         }
 
-      mn_now = rt_now;
+      mn_now = ev_rt_now;
     }
 }
 
@@ -1065,7 +1182,7 @@ ev_loop (EV_P_ int flags)
   double block;
   loop_done = flags & (EVLOOP_ONESHOT | EVLOOP_NONBLOCK) ? 1 : 0;
 
-  do
+  while (activecnt)
     {
       /* queue check watchers (and execute them) */
       if (expect_false (preparecnt))
@@ -1083,7 +1200,7 @@ ev_loop (EV_P_ int flags)
 
       /* calculate blocking time */
 
-      /* we only need this for !monotonic clockor timers, but as we basically
+      /* we only need this for !monotonic clock or timers, but as we basically
          always have timers, we just calculate it always */
 #if EV_USE_MONOTONIC
       if (expect_true (have_monotonic))
@@ -1091,8 +1208,8 @@ ev_loop (EV_P_ int flags)
       else
 #endif
         {
-          rt_now = ev_time ();
-          mn_now = rt_now;
+          ev_rt_now = ev_time ();
+          mn_now    = ev_rt_now;
         }
 
       if (flags & EVLOOP_NONBLOCK || idlecnt)
@@ -1107,26 +1224,30 @@ ev_loop (EV_P_ int flags)
               if (block > to) block = to;
             }
 
+#if EV_PERIODICS
           if (periodiccnt)
             {
-              ev_tstamp to = ((WT)periodics [0])->at - rt_now + method_fudge;
+              ev_tstamp to = ((WT)periodics [0])->at - ev_rt_now + method_fudge;
               if (block > to) block = to;
             }
+#endif
 
           if (block < 0.) block = 0.;
         }
 
       method_poll (EV_A_ block);
 
-      /* update rt_now, do magic */
+      /* update ev_rt_now, do magic */
       time_update (EV_A);
 
       /* queue pending timers and reschedule them */
       timers_reify (EV_A); /* relative timers called last */
+#if EV_PERIODICS
       periodics_reify (EV_A); /* absolute timers called first */
+#endif
 
       /* queue idle watchers unless io or timers are pending */
-      if (!pendingcnt)
+      if (idlecnt && !any_pending (EV_A))
         queue_events (EV_A_ (W *)idles, idlecnt, EV_IDLE);
 
       /* queue check watchers, to be executed first */
@@ -1134,8 +1255,10 @@ ev_loop (EV_P_ int flags)
         queue_events (EV_A_ (W *)checks, checkcnt, EV_CHECK);
 
       call_pending (EV_A);
+
+      if (loop_done)
+        break;
     }
-  while (activecnt && !loop_done);
 
   if (loop_done != 2)
     loop_done = 0;
@@ -1224,6 +1347,8 @@ ev_io_stop (EV_P_ struct ev_io *w)
   if (!ev_is_active (w))
     return;
 
+  assert (("ev_io_start called with illegal fd (must stay constant after start!)", w->fd >= 0 && w->fd < anfdmax));
+
   wlist_del ((WL *)&anfds[w->fd].head, (WL)w);
   ev_stop (EV_A_ (W)w);
 
@@ -1241,7 +1366,7 @@ ev_timer_start (EV_P_ struct ev_timer *w)
   assert (("ev_timer_start called with negative timer repeat value", w->repeat >= 0.));
 
   ev_start (EV_A_ (W)w, ++timercnt);
-  array_needsize (struct ev_timer *, timers, timermax, timercnt, (void));
+  array_needsize (struct ev_timer *, timers, timermax, timercnt, EMPTY2);
   timers [timercnt - 1] = w;
   upheap ((WT *)timers, timercnt - 1);
 
@@ -1260,10 +1385,10 @@ ev_timer_stop (EV_P_ struct ev_timer *w)
   if (((W)w)->active < timercnt--)
     {
       timers [((W)w)->active - 1] = timers [timercnt];
-      downheap ((WT *)timers, timercnt, ((W)w)->active - 1);
+      adjustheap ((WT *)timers, timercnt, ((W)w)->active - 1);
     }
 
-  ((WT)w)->at = w->repeat;
+  ((WT)w)->at -= mn_now;
 
   ev_stop (EV_A_ (W)w);
 }
@@ -1276,29 +1401,36 @@ ev_timer_again (EV_P_ struct ev_timer *w)
       if (w->repeat)
         {
           ((WT)w)->at = mn_now + w->repeat;
-          downheap ((WT *)timers, timercnt, ((W)w)->active - 1);
+          adjustheap ((WT *)timers, timercnt, ((W)w)->active - 1);
         }
       else
         ev_timer_stop (EV_A_ w);
     }
   else if (w->repeat)
-    ev_timer_start (EV_A_ w);
+    {
+      w->at = w->repeat;
+      ev_timer_start (EV_A_ w);
+    }
 }
 
+#if EV_PERIODICS
 void
 ev_periodic_start (EV_P_ struct ev_periodic *w)
 {
   if (ev_is_active (w))
     return;
 
-  assert (("ev_periodic_start called with negative interval value", w->interval >= 0.));
-
-  /* this formula differs from the one in periodic_reify because we do not always round up */
-  if (w->interval)
-    ((WT)w)->at += ceil ((rt_now - ((WT)w)->at) / w->interval) * w->interval;
+  if (w->reschedule_cb)
+    ((WT)w)->at = w->reschedule_cb (w, ev_rt_now);
+  else if (w->interval)
+    {
+      assert (("ev_periodic_start called with negative interval value", w->interval >= 0.));
+      /* this formula differs from the one in periodic_reify because we do not always round up */
+      ((WT)w)->at += ceil ((ev_rt_now - ((WT)w)->at) / w->interval) * w->interval;
+    }
 
   ev_start (EV_A_ (W)w, ++periodiccnt);
-  array_needsize (struct ev_periodic *, periodics, periodicmax, periodiccnt, (void));
+  array_needsize (struct ev_periodic *, periodics, periodicmax, periodiccnt, EMPTY2);
   periodics [periodiccnt - 1] = w;
   upheap ((WT *)periodics, periodiccnt - 1);
 
@@ -1317,12 +1449,21 @@ ev_periodic_stop (EV_P_ struct ev_periodic *w)
   if (((W)w)->active < periodiccnt--)
     {
       periodics [((W)w)->active - 1] = periodics [periodiccnt];
-      downheap ((WT *)periodics, periodiccnt, ((W)w)->active - 1);
+      adjustheap ((WT *)periodics, periodiccnt, ((W)w)->active - 1);
     }
 
   ev_stop (EV_A_ (W)w);
 }
 
+void
+ev_periodic_again (EV_P_ struct ev_periodic *w)
+{
+  /* TODO: use adjustheap and recalculation */
+  ev_periodic_stop (EV_A_ w);
+  ev_periodic_start (EV_A_ w);
+}
+#endif
+
 void
 ev_idle_start (EV_P_ struct ev_idle *w)
 {
@@ -1330,7 +1471,7 @@ ev_idle_start (EV_P_ struct ev_idle *w)
     return;
 
   ev_start (EV_A_ (W)w, ++idlecnt);
-  array_needsize (struct ev_idle *, idles, idlemax, idlecnt, (void));
+  array_needsize (struct ev_idle *, idles, idlemax, idlecnt, EMPTY2);
   idles [idlecnt - 1] = w;
 }
 
@@ -1338,7 +1479,7 @@ void
 ev_idle_stop (EV_P_ struct ev_idle *w)
 {
   ev_clear_pending (EV_A_ (W)w);
-  if (ev_is_active (w))
+  if (!ev_is_active (w))
     return;
 
   idles [((W)w)->active - 1] = idles [--idlecnt];
@@ -1352,7 +1493,7 @@ ev_prepare_start (EV_P_ struct ev_prepare *w)
     return;
 
   ev_start (EV_A_ (W)w, ++preparecnt);
-  array_needsize (struct ev_prepare *, prepares, preparemax, preparecnt, (void));
+  array_needsize (struct ev_prepare *, prepares, preparemax, preparecnt, EMPTY2);
   prepares [preparecnt - 1] = w;
 }
 
@@ -1360,7 +1501,7 @@ void
 ev_prepare_stop (EV_P_ struct ev_prepare *w)
 {
   ev_clear_pending (EV_A_ (W)w);
-  if (ev_is_active (w))
+  if (!ev_is_active (w))
     return;
 
   prepares [((W)w)->active - 1] = prepares [--preparecnt];
@@ -1374,7 +1515,7 @@ ev_check_start (EV_P_ struct ev_check *w)
     return;
 
   ev_start (EV_A_ (W)w, ++checkcnt);
-  array_needsize (struct ev_check *, checks, checkmax, checkcnt, (void));
+  array_needsize (struct ev_check *, checks, checkmax, checkcnt, EMPTY2);
   checks [checkcnt - 1] = w;
 }
 
@@ -1382,7 +1523,7 @@ void
 ev_check_stop (EV_P_ struct ev_check *w)
 {
   ev_clear_pending (EV_A_ (W)w);
-  if (ev_is_active (w))
+  if (!ev_is_active (w))
     return;
 
   checks [((W)w)->active - 1] = checks [--checkcnt];
@@ -1397,7 +1538,7 @@ void
 ev_signal_start (EV_P_ struct ev_signal *w)
 {
 #if EV_MULTIPLICITY
-  assert (("signal watchers are only supported in the default loop", loop == default_loop));
+  assert (("signal watchers are only supported in the default loop", loop == ev_default_loop_ptr));
 #endif
   if (ev_is_active (w))
     return;
@@ -1410,7 +1551,7 @@ ev_signal_start (EV_P_ struct ev_signal *w)
 
   if (!((WL)w)->next)
     {
-#if WIN32
+#if _WIN32
       signal (w->signum, sighandler);
 #else
       struct sigaction sa;
@@ -1440,7 +1581,7 @@ void
 ev_child_start (EV_P_ struct ev_child *w)
 {
 #if EV_MULTIPLICITY
-  assert (("child watchers are only supported in the default loop", loop == default_loop));
+  assert (("child watchers are only supported in the default loop", loop == ev_default_loop_ptr));
 #endif
   if (ev_is_active (w))
     return;
@@ -1453,7 +1594,7 @@ void
 ev_child_stop (EV_P_ struct ev_child *w)
 {
   ev_clear_pending (EV_A_ (W)w);
-  if (ev_is_active (w))
+  if (!ev_is_active (w))
     return;
 
   wlist_del ((WL *)&childs [w->pid & (PID_HASHSIZE - 1)], (WL)w);
@@ -1507,14 +1648,14 @@ ev_once (EV_P_ int fd, int events, ev_tstamp timeout, void (*cb)(int revents, vo
       once->cb  = cb;
       once->arg = arg;
 
-      ev_watcher_init (&once->io, once_cb_io);
+      ev_init (&once->io, once_cb_io);
       if (fd >= 0)
         {
           ev_io_set (&once->io, fd, events);
           ev_io_start (EV_A_ &once->io);
         }
 
-      ev_watcher_init (&once->to, once_cb_to);
+      ev_init (&once->to, once_cb_to);
       if (timeout >= 0.)
         {
           ev_timer_set (&once->to, timeout, 0.);
@@ -1523,3 +1664,7 @@ ev_once (EV_P_ int fd, int events, ev_tstamp timeout, void (*cb)(int revents, vo
     }
 }
 
+#ifdef __cplusplus
+}
+#endif
+
