@@ -52,20 +52,23 @@ poll_modify (EV_P_ int fd, int oev, int nev)
 
   if (idx < 0) /* need to allocate a new pollfd */
     {
-      idx = pollcnt++;
-      array_needsize (struct pollfd, polls, pollmax, pollcnt, );
+      pollidxs [fd] = idx = pollcnt++;
+      array_needsize (struct pollfd, polls, pollmax, pollcnt, EMPTY2);
       polls [idx].fd = fd;
     }
 
+  assert (polls [idx].fd == fd);
+
   if (nev)
     polls [idx].events =
         (nev & EV_READ ? POLLIN : 0)
         | (nev & EV_WRITE ? POLLOUT : 0);
   else /* remove pollfd */
     {
-      if (idx < pollcnt--)
+      pollidxs [fd] = -1;
+
+      if (idx < --pollcnt)
         {
-          pollidxs [fd] = -1;
           polls [idx] = polls [pollcnt];
           pollidxs [polls [idx].fd] = idx;
         }
@@ -76,7 +79,7 @@ static void
 poll_poll (EV_P_ ev_tstamp timeout)
 {
   int i;
-  int res = poll (polls, pollcnt, ceil (timeout * 1000.));
+  int res = poll (polls, pollcnt, (int)ceil (timeout * 1000.));
 
   if (res < 0)
     {
@@ -91,12 +94,15 @@ poll_poll (EV_P_ ev_tstamp timeout)
     }
 
   for (i = 0; i < pollcnt; ++i)
-    fd_event (
-      EV_A_
-      polls [i].fd,
-      (polls [i].revents & (POLLOUT | POLLERR | POLLHUP) ? EV_WRITE : 0)
-      | (polls [i].revents & (POLLIN | POLLERR | POLLHUP) ? EV_READ : 0)
-    );
+    if (polls [i].revents & POLLNVAL)
+      fd_kill (EV_A_ polls [i].fd);
+    else
+      fd_event (
+        EV_A_
+        polls [i].fd,
+        (polls [i].revents & (POLLOUT | POLLERR | POLLHUP) ? EV_WRITE : 0)
+        | (polls [i].revents & (POLLIN | POLLERR | POLLHUP) ? EV_READ : 0)
+      );
 }
 
 static int
